import { rollup } from 'rollup';
import { terser } from 'rollup-plugin-terser';
import gulp from 'gulp';
import jest from 'gulp-jest';
import del from 'del';
import resolve from 'rollup-plugin-node-resolve';
import trainman from 'trainman';
import minimist from 'minimist';
import pkg from './package.json';
import jestConfig from './jest.config';

const options = minimist( process.argv.slice( 2 ) );
const environment = options.environment || options.e || 'acceptance';
const remote = options.remote ||
  options.r ||
  process.argv.indexOf( 'deploy' ) > 0;
const config = {
  _package: pkg,
  environmentName: environment,
  manifest: {},
  publicDir: './dist'
};
const plugins = [
  environment === 'production' && terser(),
  resolve()
];
const train = trainman( environment, config );

// test function
const testSuite = () => {
  process.env.NODE_ENV = 'test';

  return new Promise( ( resolve, reject ) => {
    const testStream = gulp.src( './test' ).
      pipe( jest( jestConfig ) );

    testStream.
      on( 'finish', () => {
        resolve();
      } );
    testStream.
      on( 'error', ( error ) => {
        reject( error );
      } );
  } );
};

// build functions
const browserBuild = async() => {
  const browserBundle = await rollup( {
    input: 'src/indexes/browser.js',
    plugins
  } );

  await browserBundle.write( {
    file: remote ?
      ( pkg.browser.split( '.js' )[ 0 ] + `-${pkg.version}.js` ) :
      pkg.browser,
    format: 'iife',
    name: 'Unimatrix'
  } );
};
const nodeBuild = async() => {
  const nodeBundle = await rollup( {
    input: 'src/indexes/node.js',
    plugins
  } );

  await nodeBundle.write( {
    file: remote ?
      ( pkg.main.split( '.js' )[0] + `-${pkg.version}.js` ) :
      pkg.main,
    format: 'cjs'
  } );
};
const cleanBuild = async() => {
  await del( [ ' ./dist/**' ] );
  await Promise.all( [ browserBuild(), nodeBuild() ] );
};

// gulp tasks
gulp.task( 'browser', browserBuild );
gulp.task( 'node', nodeBuild );
gulp.task( 'watch', [ 'build' ], () => {
  const functions = './src/functions/**';
  const browser = './src/services/browser/**';
  const node = './src/services/node/**';
  const shared = './src/services/shared/**';
  const sharedIndex = './src/indexes/shared.js';
  const nodeIndex = './src/indexes/node.js';
  const browserIndex = './src/indexes/browser.js';
  const unimatrix = './src/unimatrix.js';

  // watch shared
  gulp.watch( [
    functions,
    shared,
    sharedIndex,
    unimatrix
  ],
  [ 'browser', 'node' ]
  );
  // watch browser
  gulp.watch(
    [
      browser,
      browserIndex
    ],
    [ 'browser' ]
  );
  // watch node
  gulp.watch(
    [
      node,
      nodeIndex
    ],
    [ 'node' ]
  );
} );
gulp.task( 'build', async() => {
  await cleanBuild();

  if ( remote ) {
    await testSuite();
  }
} );
gulp.task( 'test', async() => {
  await testSuite();
} );

train.installTask( 'deploy' );
