let injector;
const path = require( 'path' );
const fs = require( 'fs-extra' );

try {
  injector = require( '../dist/unimatrix-sdk-node' );
}
catch ( _error ) {
  const version = require( '../package.json' ).version;

  injector = require( `../dist/unimatrix-sdk-node-${version}` );
}

const snakeCase = function( string ) {
  return ( string || '' ).
    replace( /^\s+|\s+$/gm, '' ).
    replace( /([^A-Z][A-Z])|([^\s-][\s-]+[^\s-])/g, function( match ) {
      return match.charAt( 0 ) + '_' + match.slice( -1 );
    } ).
    replace( /[_]+/g, '_' ).
    toLowerCase();
};
const resourceList = injector().
  list( 'Resources' );
const getAttributes = function( resourceName ) {
  const attributes = injector().
    get( `Resources/${resourceName}` ).
    getAttributes();
  const attributeList = attributes.map( ( attribute ) => {
    const { name, alias, type, persist } = attribute;

    return { name, alias, type, persist };
  } );

  return {
    name: resourceName,
    attributes: attributeList
  };
};
const getTests = async function() {
  try {
    const currentTests = await fs.readdir(
      path.join( __dirname, '../', 'test/unit/services/resources' )
    );

    return currentTests.
      reduce( ( acc, test ) => {
        acc[ test ] = true;

        return acc;
      }, {} );
  }
  catch ( error ) {
    console.error( `directory read error: ${error}` );
    process.exit( 1 );
  }
};
const filterResources = async function( resources ) {
  const tests = await getTests();

  return resources.filter( resource => !tests[ snakeCase( resource ) ] );
};
const generateTest = async function( resource ) {
  const name = snakeCase( resource.name );
  const writeFile = async function( name, file, body ) {
    try {
      await fs.outputFile(
        path.join(
          __dirname, '../', `test/unit/services/resources/${name}/${file}.js`
        ),
        body
      );

      return { error: false, body: `saved ${name} ${file}` };
    }
    catch ( error ) {
      return { error: true, body: `failed to save ${name} ${file}: ${error}` };
    }
  };
  const attributeData = `module.exports = ${
    JSON.stringify( resource.attributes, null, 2 ).
      replace( /"/g, "'" )
  };
`;
  const index = `
/* eslint-disable max-len */
const attributeData = require( './attributeData' );
const resourceTest = require( '../resourceTestBase' );

module.exports = function( resource ) {
  describe( '${resource.name} resource', () => {
    resourceTest( resource, attributeData );
  } );
};
`;
  const nodeTest = `
/**
* @jest-environment node
*/

import { injector } from '../../../../nodeInit';

const testIndex = require( './index' );

const resource = injector().
  get( 'Resources/${resource.name}' );

testIndex( resource );
`;
  const browserTest = `
/**
* @jest-environment jsdom
*/

import '../../../../browserInit';

const testIndex = require( './index' );
const resource = window.
  unimatrix.
  injector().
  get( 'Resources/${resource.name}' );

testIndex( resource );
`;

  return Promise.all( [
    writeFile( name, 'index', index ),
    writeFile( name, 'node.test', nodeTest ),
    writeFile( name, 'browser.test', browserTest ),
    writeFile( name, 'attributeData', attributeData )
  ] );
};

const testBuilder = async function( resources ) {
  const filteredResources = await filterResources( resources );
  const parsedResourceAttributes = filteredResources.map( getAttributes );
  const resourcePromises = parsedResourceAttributes.map( generateTest );
  let errors = 0;

  const results = await Promise.all( resourcePromises );

  results.forEach( result => {
    result.forEach( message => {
      if ( message.error ) {
        errors++;

        return console.error( message.body );
      }

      console.info( message.body );
    } );
  } );
  errors ?
    console.error( `test building completed with ${errors} errors` ) :
    console.info( 'test building completed successfully' );
};

testBuilder( resourceList );
