import { isFunction, isArrayLike, isDefined, isString } from './rtti';

function each( object, iterator, context ) {
  var key;
  var length;

  if ( object ) {
    if ( isFunction( object ) ) {
      for ( key in object ) {
        // note: on IE8, the result of querySelectorAll is an object without
        //       a hasOwnProperty function
        if ( key !== 'prototype' &&
             key !== 'length' && key !== 'name' &&
             ( !object.hasOwnProperty || object.hasOwnProperty( key ) ) ) {
          if ( iterator.call( context, object[ key ], key ) === false ) {
            return object;
          }
        }
      }
    }
    else if ( isArrayLike( object ) ) {
      for ( key = 0, length = object.length; key < length; key++ ) {
        if ( iterator.call( context, object[ key ], key ) === false ) {
          return object;
        }
      }
    }
    else {
      for ( key in object ) {
        if ( object.hasOwnProperty( key ) ) {
          if ( iterator.call( context, object[ key ], key ) === false ) {
            return object;
          }
        }
      }
    }
  }
  return object;
}

function map( object, iterator, context ) {
  var result;
  var key;
  var length;

  if ( object ) {
    if ( isFunction( object ) ) {
      throw new Error( 'A function cannot be mapped.' );
    }
    else if ( isArrayLike( object ) ) {
      result = [];
      for ( key = 0, length = object.length; key < length; key++ ) {
        result.push( iterator.call( context, object[ key ], key ) );
      }
    }
    else {
      result = {};
      for ( key in object ) {
        if ( object.hasOwnProperty( key ) ) {
          result[ key ] = iterator.call( context, object[ key ], key );
        }
      }
    }
  }
  return result;
}

function detect( array, comparator ) {

  if ( !array ) {
    return undefined;
  }
  if ( !isDefined( array.length ) ) {
    throw 'dtect: An array argument is required';
  }
  if ( !isFunction( comparator ) ) {
    throw 'detect: ' +
      'A comparator function argument is required';
  }

  for ( var index = 0; index < array.length; index++ ) {
    var item = array[ index ];

    if ( comparator( item ) ) {
      return item;
    }
  }

  return undefined;
}

function index( array, fieldOrIndexor ) {

  if ( !array ) {
    return undefined;
  }
  if ( !isDefined( array.length ) ) {
    throw 'index: An array argument is required';
  }
  if ( !isString( fieldOrIndexor ) &&
    !isFunction( fieldOrIndexor ) ) {
    throw 'index: ' +
      'A field name or indexor function argument is required';
  }

  var result = {};
  for ( var index = 0; index < array.length; index++ ) {
    var item = array[ index ];
    result[ item[ fieldOrIndexor ] ] = item;
  }

  return result;
}

function chunk( array, increment ) {
  var result = [];
  var position = increment;


  while ( position - increment < array.length ) {
    result.push( array.slice( position - increment, position ) );

    position += increment;
  }

  return result;
}

export {
  each,
  map,
  detect,
  index,
  chunk
};