import { isString, isNumber } from '../../functions/rtti';
import { each } from '../../functions/enumeration';
import { extend } from '../../functions/object';
import { error } from '../../functions/error';
import { module } from '../../unimatrix';

// .name Unimatrix/Http
//
// .description
// The +Unimatrix/Http+ provides implements an asynchronous http ( AJAX )
// factory.
//
//
//  requestConfiguration:
//
//  method - {string}
//  url - {string}
//  params - {object}
//  data - {string|object}
//  headers – {Object}

module( 'Unimatrix' ).
  service(
    'Http',
    [
      'Unimatrix/UserAgent', 'Unimatrix/Q', 'Unimatrix/URI',
      function( UserAgent, Q, URI ) {

        var userAgent = UserAgent();

        function isSuccess( status ) {
          return 200 <= status && status < 300;
        }

        function buildUrl( url, params ) {

          if ( typeof params !== 'object' ||
            !Object.keys( params ).length ) {
            return url;
          }

          return url +
            ( ( url.indexOf( '?' ) === -1 ) ? '?' : '&' ) +
            URI.queryFromObject( params );

        }

        function Http( config ) {
          config = config || {};

          var deferred = Q.deferred();
          var promise = deferred.promise;

          var requestMethod = ( config.method || 'get' ).toUpperCase();
          var requestUrl = buildUrl( config.url, config.params ) || '';
          var requestBody = config.data ?
            ( isString( config.data ) ?
              config.data :
              JSON.stringify( config.data ) ) :
            null;
          var httpRequest;

          if ( !( userAgent.name() === 'internet explorer' &&
                  userAgent.version()[ 0 ] === '9' ) ) {
            if ( window.XMLHttpRequest ) { // Mozilla, Safari, ...
              httpRequest = new XMLHttpRequest();
            }
            else if ( window.ActiveXObject ) { // IE
              try {
                httpRequest = new ActiveXObject( 'Msxml2.XMLHTTP' );
              }
              catch ( e ) {
                try {
                  httpRequest = new ActiveXObject( 'Microsoft.XMLHTTP' );
                }
                catch ( error ) {
                  console.error( error );
                }
              }
            }
          }
          else {
            httpRequest = new XDomainRequest();
          }

          if ( !httpRequest ) {
            throw error(
              'The user agent lacks an HTTP request implementaiton.'
            );
          }

          httpRequest.onreadystatechange = function() {
            // call the .then callback/fufill the promise
            try {
              if ( httpRequest.readyState === 4 ) {
                var status = Math.max( httpRequest.status, 0 );
                ( isSuccess( status ) ? deferred.resolve : deferred.reject )(
                  {
                    'status':     status,
                    'statusText': httpRequest.statusText,
                    'headers':    httpRequest.getAllResponseHeaders(),
                    'body':       httpRequest.responseText
                  }
                );
              }
            }
            catch ( error ) {
              console.error( error );
            }
          };

          // IE9 only
          httpRequest.onload = function() {
            deferred.resolve(
              {
                'status':     200,
                'statusText': 'OK',
                'headers':    {},
                'body':       httpRequest.responseText
              }
            );
          };

          httpRequest.onerror = function() {
            deferred.reject(
              {
                'status':     500,
                'statusText': 'Unexpected Error',
                'headers':    {},
                'body':       httpRequest.responseText
              }
            );
          };

          httpRequest.ontimeout = function() {
            deferred.reject(
              {
                'status':     408,
                'statusText': 'Unexpected Timeout Error',
                'headers':    {},
                'body':       httpRequest.responseText
              }
            );
          };
          // parse url (HTTP access control?)
          httpRequest.open( requestMethod, requestUrl, true );

          if ( requestBody ) {
            httpRequest.setRequestHeader( 'Content-Type', 'application/json' );
          }

          // IE Documentation: The timeout property may be set only in the time
          // interval between a call to the open method and the first call to
          // the send method.
          if ( isNumber( config.timeout ) && config.timeout >= 0 ) {
            httpRequest.timeout = config.timeout;
          }
          httpRequest.send( requestBody );

          return promise;
        }

        each(
          [ 'get', 'head', 'delete' ],
          function( method ) {
            Http[ method ] = function( url, requestConfiguration ) {
              return Http(
                extend(
                  { 'method': method, 'url': url },
                  requestConfiguration || {}
                )
              );
            };
          }
        );
        each(
          [ 'post', 'put', 'patch' ],
          function( method ) {
            Http[ method ] = function( url, body, requestConfiguration ) {
              return Http(
                extend(
                  { 'method': method, 'body': body },
                  requestConfiguration || {}
                )
              );
            };
          }
        );

        return Http;
      }
    ]
  );