import { module } from '../../unimatrix';
import { each } from '../../functions/enumeration';
import { extend } from '../../functions/object';
import { error } from '../../functions/error';
import { defer } from '../../functions/defer';

module( 'Unimatrix' ).
  service(
    'Notification',
    [
      function() {

        //---------------------------------------------------------------------
        // implementation

        function notifySubscribers( typeName, notification, subscribers ) {
          each(
            subscribers,
            function( subscriber ) {
              // note: if a subscriber callback returns 'false' the
              //       notification propagation is halted.
              return subscriber.callback.apply(
                subscriber.callbackContext || null,
                [
                  extend( {}, notification || {}, { type: typeName } ),
                  subscriber.payload
                ]
              );
            }
          );
        }

        //---------------------------------------------------------------------
        // instance initialization

        function Notification( typeNames ) {
          this.subscribers = {};
          each(
            typeNames,
            function( typeName ) {
              this.subscribers[ typeName ] = [];
            },
            this
          );
          this.delegates = [];


          // A notification center with a lock count greater then 0 will not
          // dispatch or send notifcations.
          this.$$lockCount = 0;
        }

        //---------------------------------------------------------------------
        // instance operations

        Notification.prototype = {

          'constructor': Notification,

          subscribe: function( typeName, callback, callbackContext, payload ) {
            var subscriberList = this.subscribers[ typeName ];
            if ( !subscriberList ) {
              return;
            }
            this.unsubscribe( typeName, callback );
            subscriberList.push( {
              'callback': callback,
              'callbackContext': callbackContext,
              'payload': payload
            } );
          },

          unsubscribe: function( typeName, callback, callbackContext ) {
            var subscriberList = this.subscribers[ typeName ];
            if ( !subscriberList ) {
              return;
            }
            var index = subscriberList.length - 1;
            while ( index >= 0 ) {
              var subscriber = subscriberList[ index ];
              if ( subscriber.callback === callback &&
                   subscriber.callbackContext === callbackContext ) {
                subscriberList.splice( index, 1 );
                break;
              }
              index--;
            }
          },

          delegate: function( notificationCenter ) {
            if ( this.delegates.indexOf( notificationCenter ) === -1 ) {
              this.delegates.push( notificationCenter );
            }
          },

          send: function( typeName, message ) {
            var subscriberList = this.subscribers[ typeName ];
            if ( !subscriberList ) {
              throw error(
                'The event named \'' + typeName + '\' is not supported.'
              );
            }
            if ( this.$$lockCount > 0 ) {
              return;
            }
            notifySubscribers( typeName, message, subscriberList );
            each(
              this.delegates,
              function( delegate ) {
                delegate.send( typeName, message );
              }
            );
          },

          dispatch: function( typeName, message ) {
            var subscriberList = this.subscribers[ typeName ];
            if ( !subscriberList ) {
              throw error(
                'The event named \'' + typeName + '\' is not supported.'
              );
            }
            if ( this.$$lockCount > 0 ) {
              return;
            }
            defer( function() {
              notifySubscribers( typeName, message, subscriberList );
            } );
            each(
              this.delegates,
              function( delegate ) {
                delegate.dispatch( typeName, message );
              }
            );
          },

          //
          // .method lock
          //
          // The +lock+ method locks the notifcation center. A locked
          // notification center will not dispatch or send notifcations.
          //
          lock: function() {
            this.$$lockCount++;
          },

          //
          // .method unlock
          //
          // The +unlock+ method unlock the notifcation center. A locked
          // notification center will not dispatch or send notifcations.
          //
          unlock: function() {
            if ( this.$$lockCount > 0 ) {
              this.$$lockCount--;
            }
          }

        };

        return function( typeNames ) {
          return new Notification( typeNames );
        };

      }
    ]
  );
