import { module } from '../../unimatrix';
import { digest } from '../../functions/string';
import { extend, merge } from '../../functions/object';
import {
  isArray,
  isFunction,
  isObject,
  isDefined
} from '../../functions/rtti';
import { each, map, chunk } from '../../functions/enumeration';

module( 'Unimatrix' ).
  service(
    'Operation',
    [
      'Unimatrix/Q',
      'Unimatrix/Http',
      'Unimatrix/URI',
      'Unimatrix/ResourceParser',
      function( Q, Http, URI, ResourceParser ) {
        var OperationFactory = function( _url, _options ) {
          var Operation = function( url, options ) {
            this.url = url.toString();
            this.options = extend( {}, options || {} );
            this._key = undefined;
          };

          Operation.prototype = {
            'constructor': Operation,
            key: function() {
              var self = this;

              return self._key || ( this._key = function() {
                var result = 0;
                var query = URI.queryFromObject(
                  self.options[ 'params' ] || ''
                );
                if ( self.url || query.length > 0 ) {
                  query = query.split( '&' ).
                    sort().
                    join( '&' );
                  result = digest(
                    URI( self.url ).
                      joinQuery( query ).
                      toString()
                  );
                }

                return result;
              }() );
            },

            where: function( parameters ) {
              return this._spawn( { params: parameters } );
            },

            order: function( by, direction ) {
              return this._spawn(
                { params: { sort_by: by, sort_direction: direction } }
              );
            },

            include: function() {
              return this._spawn(
                {
                  params: {
                    include: this.
                      _normalizeInclude.
                      apply( this, arguments )
                  }
                }
              );
            },

            limit: function( _limit ) {
              return this._spawn( { params: { count: _limit } } );
            },

            offset: function( _offset ) {
              return this._spawn( { params: { offset: _offset } } );
            },

            find: function( id ) {
              return this._spawn(
                { params: { 'id': id }, typeOfResult: 'object' }
              );
            },

            take: function() {
              return this._spawn( { typeOfResult: 'object' } );
            },

            batch: function( _count ) {
              _count = _count || 100;

              return this._spawn( { batchSize: _count } );
            },

            then: function() {
              return this.query().
                then( arguments[0], arguments[1] );
            },

            query: function() {
              return this._request( this.options[ 'method' ] || 'GET' );
            },

            write: function( name, resources ) {
              // is the resources parameter an array?
              if ( !isArray( resources ) ) {
                // if not, assume it's a single resource
                // and wrap it in an array.
                resources = [ resources ];
              }

              var data = { name: name };
              data.resources = map(
                resources,
                function( resource ) {
                  if ( isFunction( resource.getAttributesByAlias ) ) {
                    return resource.getAttributesByAlias();
                  }
                  return resource;
                }
              );

              return this._request( this.options[ 'method' ] || 'POST', data );
            },

            destroy: function() {
              return this._request( this.options[ 'method' ] || 'DELETE' );
            },

            //-----------------------------------------------------------------
            // implementation

            _spawn: function( options ) {
              var spawn = new Operation( this.url, this.options );
              merge(
                spawn.options,
                options
              );
              return spawn;
            },

            _normalizeInclude: function() {
              var include = {};

              for ( var index = 0, length = arguments.length;
                index < length;
                index++ ) {
                var argument = arguments[ index ];
                if ( isArray( argument ) ) {
                  each(
                    argument,
                    function( value ) {
                      merge(
                        include,
                        this._normalizeInclude( value )
                      );
                    },
                    this
                  );
                }
                else if ( isObject( argument ) ) {
                  each(
                    argument,
                    function( value, key ) {
                      if ( !include[ key ] || include[ key ] === true ) {
                        include[ key ] = this._normalizeInclude( value );
                      }
                      else {
                        merge(
                          include[ key ],
                          this._normalizeInclude( value )
                        );
                      }
                    },
                    this
                  );
                }
                else {
                  include[ argument ] = true;
                }
              }
              return include;
            },

            _request: function( method, data ) {
              var self = this;
              var options = self.options;
              var configuration = {};
              var baseUri = isDefined( OperationFactory.baseUrl ) ?
                URI( OperationFactory.baseUrl ) : undefined;
              var httpErrorback = function( response ) {
                return Q.reject( response );
              };

              // base config
              configuration.params = extend(
                {},
                OperationFactory.defaultParameters,
                options.params || {}
              );
              configuration.method = method;
              if ( isDefined( options.localUri ) ) {
                baseUri = URI( options.localUri );
              }
              configuration.url = isDefined( baseUri ) ?
                baseUri.joinPath( self.url ).
                  toString() : self.url;
              configuration.headers = self.options[ 'headers' ];

              // without batching
              if ( !options.batchSize ) {
                configuration.params = extend(
                  {},
                  OperationFactory.defaultParameters,
                  options.params || {}
                );
                if ( data ) {
                  var requestData = {};

                  requestData[ data.name ] = data.resources;
                  configuration.data = requestData;
                }
                configuration.timeout = this.options.timeout;
                return Http( configuration ).
                  then(
                    function( response ) {
                      var result = response;
                      var data = JSON.parse( response.body );
                      if ( data && isObject( data ) ) {
                        var responseParser = ResourceParser( data );
                        result = responseParser.getResources();
                        if ( isArray( result ) &&
                              options[ 'typeOfResult' ] === 'object' ) {
                          result = result[ 0 ];
                        }
                        else {
                          var unlimitedCount = responseParser.
                            responseMeta.
                            unlimited_count;

                          result.count = responseParser.
                            responseMeta.
                            count;

                          if ( isDefined( unlimitedCount ) ) {
                            result.unlimitedCount = unlimitedCount;
                          }
                        }
                      }
                      return result;
                    },
                    httpErrorback
                  );
              }

              // with batching
              return this._batch( configuration, data );
            },

            _batch: function( configuration, data ) {
              var self = this;
              var options = self.options;

              if ( !data ) {
                var config = extend( {}, configuration );
                var batchOptions = {
                  offset: configuration.params.offset || 0,
                  limit: configuration.params.count ?
                    options.params.count : null,
                  count: options.batchSize,
                  total: 0,
                  final: false
                };

                if ( batchOptions.limit === options.batchSize ) {
                  batchOptions.final = true;
                }

                config.params.count = batchOptions.count;

                return this._getLoop( config, batchOptions, [] );
              }

              data.resources = chunk( data.resources, options.batchSize );

              return this._postLoop( configuration, data, [] );
            },
            _postLoop: function _postLoop( htmlConfig, currentData, results ) {
              var newConfig = extend(
                {},
                htmlConfig
              );
              var requestData = {};

              requestData[ currentData.name ] = currentData.resources[0];
              newConfig.data = requestData;
              return Http( newConfig ).
                then(
                  function( response ) {
                    var result = response;
                    var data = JSON.parse( response.body );
                    if ( data && isObject( data ) ) {
                      var responseParser = ResourceParser( data );
                      result = responseParser.getResources();
                    }

                    var newResults = results.concat( result );

                    if ( currentData.resources.length <= 1 ) {
                      return newResults;
                    }
                    var newData = {
                      name: currentData.name,
                      resources: currentData.resources.slice( 1 )
                    };

                    return _postLoop( newConfig, newData, newResults );
                  },
                  function( error ) {
                    if ( !error.remaining ) {
                      error.remaining = currentData.
                        resources.
                        reduce( function( accumulator, item ) {
                          return accumulator.concat( item );
                        }, [] );
                    }

                    return Q.reject( error );
                  }
                );
            },
            _getLoop: function _getLoop( htmlConfig, loopOptions, results ) {
              return Http( htmlConfig ).
                then(
                  function( response ) {
                    var result = response;
                    var data = JSON.parse( response.body );
                    var newConfig = extend(
                      {},
                      htmlConfig
                    );
                    if ( data && isObject( data ) ) {
                      var responseParser = ResourceParser( data );
                      result = responseParser.getResources();
                      loopOptions.total = responseParser.
                        responseMeta.
                        unlimited_count || loopOptions.total;
                    }
                    var newResults = results.concat( result );

                    if ( loopOptions.limit ) {
                      if ( loopOptions.final ) {
                        return newResults;
                      }
                      if ( newResults.length + loopOptions.count >=
                        loopOptions.limit ) {
                        var finalLimit = loopOptions.limit -
                          newResults.length;

                        loopOptions.final = true;
                        newConfig.params.count = finalLimit;
                      }
                    }
                    loopOptions.offset += loopOptions.count;
                    if ( loopOptions.offset < loopOptions.total ) {
                      newConfig.params.offset = loopOptions.offset;

                      return _getLoop(
                        newConfig,
                        loopOptions,
                        newResults
                      );
                    }

                    return newResults;
                  },
                  function( response ) {
                    return Q.reject( response );
                  }
                );
            }
          };

          return new Operation( _url, _options );
        };

        //---------------------------------------------------------------------
        // initialization

        OperationFactory.baseUrl = null;
        OperationFactory.defaultParameters = {};

        return OperationFactory;
      }
    ]
  );
