import { module, injector } from '../../../unimatrix';
import { camelCase, pluralize } from '../../../functions/string';
import { isArray, isFunction, isObject } from '../../../functions/rtti';
import { each, index, map, detect } from '../../../functions/enumeration';
import { extend } from '../../../functions/object';
import { error } from '../../../functions/error';

module( 'Unimatrix' ).
  service(
    'ResourceParser',
    [
      function() {
        var ResourceParser = function( response ) {
          var self = this;

          this.responseMeta = extend(
            {},
            response[ '$this' ] || {}
          );
          this.resources = {};
          this.indexedResources = {};
          this.associations = extend(
            {},
            response[ '$associations' ] || {}
          );

          for ( var key in response ) {
            if ( key.charAt( 0 ) !== '$' ) {
              var value = response[ key ];
              if ( isArray( value ) ) {
                this.resources[ key ] = value;
                this.indexedResources[ key ] = index( value, 'id' );
              }
            }
          }

          this.getName = function() {
            return this.responseMeta[ 'name' ];
          };

          this.getTypeName = function() {
            return this.responseMeta[ 'type_name' ];
          };

          this.getCount = function() {
            return this.responseMeta[ 'count' ];
          };

          this.getUnlimitedCount = function() {
            return this.responseMeta[ 'unlimited_count' ];
          };

          this.getResources = function() {
            var result = [];
            var resources;
            var resourcesName = this.responseMeta[ 'name' ];
            var resourcesTypeName = this.responseMeta[ 'type_name' ];
            var resourceIds = this.responseMeta[ 'ids' ];

            if ( isArray( resourceIds ) && resourceIds[ 0 ] ) {
              resources = this.indexedResources[ resourcesName ];
              each(
                resourceIds,
                function( id ) {
                  var resource = resources[ id ];
                  if ( resource ) {
                    result.push( self._resolveResource(
                      resource,
                      { 'name': resourcesName, 'type_name': resourcesTypeName }
                    ) );
                  }
                }
              );
            }
            else {
              resources = this.resources[ resourcesName ];
              each(
                resources,
                function( resource ) {
                  result.push( self._resolveResource(
                    resource,
                    { 'name': resourcesName, 'type_name': resourcesTypeName }
                  ) );
                }
              );
            }
            return result;
          };

          this._resolveResource = function( attributes, options ) {
            options = options || {};
            if ( !isObject( attributes ) ) {
              return null;
            }

            var associations = null;
            var provider = null;
            var result = null;
            var realmUuid = attributes[ 'realm_uuid' ];
            var typeName = attributes[ 'type_name' ] || options[ 'type_name' ];
            var name = options[ 'name' ] || pluralize( typeName );
            var providerName = camelCase( typeName );

            if ( providerName ) {
              provider = injector().
                get( 'Resources/' + providerName );
            }
            if ( !provider ) {
              if ( options[ 'type_name' ] ) {
                providerName = camelCase( options[ 'type_name' ] );
                provider = injector().
                  get( 'Resources/' + providerName );
                if ( !provider ) {
                  throw error( providerName, 'is an unknown resource.' );
                }
                if ( isFunction( provider.resourceFromBlueprint ) ) {
                  provider = provider.
                    resourceFromBlueprint( realmUuid, typeName );
                }
              }
            }
            if ( provider ) {
              associations = this.
                _resolveResourceAssociations( name, attributes[ 'id' ] );
              result = new provider( attributes, associations );
            }
            return result;
          };

          this._resolveResourceAssociations = function( name, id ) {
            var result = {};
            var associations = this.associations[ name ];
            if ( associations ) {
              var associationsByType = detect(
                associations,
                function( associate ) {
                  return associate.id === id;
                }
              );
              for ( var key in associationsByType ) {
                var associatedResources = this.indexedResources[ key ];
                if ( associatedResources ) {
                  var association = associationsByType[ key ];
                  var typeName = association.type_name;
                  var ids = association.ids;
                  result[ key ] = map(
                    ids || [],
                    function( associatedId ) {
                      return self._resolveResource(
                        associatedResources[ associatedId ],
                        { 'name': key, 'type_name': typeName }
                      );
                    }
                  );
                }
              }
            }
            return result;
          };

        };

        return function( response ) {
          return new ResourceParser( response );
        };
      }
    ]
  );
