import { each } from '../../functions/enumeration';
import { error } from '../../functions/error';
import {
  isArray,
  isObject,
  isDefined,
  isString,
  isFunction,
  isUndefined
} from '../../functions/rtti';
import { extend } from '../../functions/object';
import { lowerCase } from '../../functions/string';
import { module } from '../../unimatrix';

module( 'Unimatrix' ).
  service(
    'URI',
    function() {

      //-----------------------------------------------------------------------
      // constants

      var URI_PARSER = /^(?:((?:https?|s?ftp)):)?(?:(?:\/\/)?((?:[^:/\s]+[.][^:/\s]+)+|localhost))?(?::(\d*))?(?:\/?([^\s?#]+))?(?:[?]([^?#]*))?(?:[#](.*)?)?/;
      var PATH_PARSER = /\/?([^\s?#]+)/;

      //-----------------------------------------------------------------------
      // implementation

      var normalizePath = function( path ) {
        var result = null;
        if ( path ) {
          var pathSegments = PATH_PARSER.exec( path );
          if ( pathSegments ) {
            result = pathSegments[ 1 ].replace( /\/+$/, '' );
          }
        }
        return result;
      };

      function assembleQueryParameter( key, value ) {

        var results = [];

        if ( isArray( value ) ) {
          each( value, function( arrayValue ) {
            results = results.concat(
              assembleQueryParameter( key + '[]', arrayValue )
            );
          } );
        }
        else if ( isObject( value ) ) {
          each( value, function( objectValue, objectKey ) {
            results = results.concat(
              assembleQueryParameter(
                key + '%5B' + objectKey + '%5D',
                objectValue
              )
            );
          } );
        }
        else if ( value ) {
          results.push( key + '=' + encodeQuery( value ) );
        }

        return results;

      }

      //-----------------------------------------------------------------------
      // implementation

      var encodeQuery = function( value, percentEncodeSpaces ) {
        return encodeURIComponent( value ).
          replace( /%40/gi, '@' ).
          replace( /%3A/gi, ':' ).
          replace( /%24/g, '$' ).
          replace( /%2C/gi, ',' ).
          replace( /%20/g, ( percentEncodeSpaces ? '%20' : '+' ) );
      };

      var encodeQuerySegment = function( value ) {
        return encodeQuery( value, true ).
          replace( /%26/gi, '&' ).
          replace( /%3D/gi, '=' ).
          replace( /%2B/gi, '+' );
      };

      var queryFromObject = function( object ) {
        var results = [];
        each( object, function( values, key ) {
          results = results.concat( assembleQueryParameter( key, values ) );
        } );

        return results.join( '&' );
      };

      var objectFromQuery = function( query ) {
        query = query || '';
        var result = {};
        var pairs = query.toString().
          split( '&' );
        each(
          pairs,
          function( pair ) {
            pair = pair.split( '=' );
            result[ pair[0] ] = pair[ 1 ] ?
              decodeURIComponent( pair[ 1 ] ) : '';
          }
        );
        return result;
      };

      var URI = function( inputUrl ) {

        var url = inputUrl ? inputUrl.toString() : null;

        this.protocol = null;
        this.hostname = null;
        this.path = null;
        this.query = null;
        this.hash = null;

        this.absolute = undefined;

        if ( url ) {

          var segments = URI_PARSER.exec( url );
          if ( !segments ) {
            throw error( 'An unrecognized uri string was given.' );
          }

          this.protocol = segments[ 1 ];
          this.hostname = segments[ 2 ];
          this.port = segments[ 3 ];
          this.path = normalizePath( segments[ 4 ] );
          this.query = segments[ 5 ];
          this.hash = segments[ 6 ];

          this.absolute = isDefined( this.hostname );
        }

        return this;

      };

      URI.prototype = {

        'constructor': URI,

        'spawn': function( block ) {

          var uri = new URI();

          uri.protocol = this.protocol;
          uri.hostname = this.hostname;
          uri.port = this.port;
          uri.path = this.path;
          uri.query = this.query;
          uri.hash = this.hash;

          if ( isFunction( block ) ) {
            block.call( null, uri );
          }

          return uri;
        },

        //-----------------------------------------------------------------
        // queries

        isAbsolute: function() {
          return this.absolute;
        },

        isRelative: function() {
          return !this.absolute;
        },

        getProtocol: function() {
          return this.protocol;
        },

        getHost: function() {
          var host = null;
          if ( this.hostname ) {
            host = this.hostname +
              ( this.port ? ( ':' + this.port ) : '' );
          }
          return host;
        },

        getHostname: function() {
          return this.hostname;
        },

        getPort: function() {
          return this.port;
        },

        getPath: function() {
          return this.path;
        },

        getFilename: function() {
          if ( !isString( this.path ) ) {
            return undefined;
          }
          var path = this.path;
          return path.substring( path.lastIndexOf( '/' ) + 1, path.length );
        },

        getFilenameExtension: function() {
          var filename = this.getFilename();
          if ( isUndefined( filename ) ) {
            return undefined;
          }
          var segments = filename.split( '.' );
          if ( segments.length === 1 ||
              ( segments[ 0 ] === '' && segments.length === 2 ) ) {
            return undefined;
          }
          return lowerCase( segments[ segments.length - 1 ] );
        },

        getQuery: function() {
          return this.query;
        },

        getHash: function() {
          return this.hash;
        },

        toString: function() {

          var result = '';
          if ( this.protocol ) {
            result += this.protocol + ':';
          }
          var host = this.getHost();
          if ( host ) {
            result += '//' + host;
          }
          else {
            // if there is no host, remove the protocol from the result
            result = '';
          }
          if ( this.path ) {
            result += '/' + normalizePath( this.path );
          }
          if ( this.query ) {
            result += '?' + this.query;
          }
          if ( this.hash ) {
            result += '#' + this.hash;
          }
          return result;
        },

        //-----------------------------------------------------------------
        // operations

        setProtocol: function( protocol ) {
          return this.spawn( function( uri ) {
            uri.protocol = protocol;
          } );
        },

        setHostname: function( hostname ) {
          return this.spawn( function( uri ) {
            if ( hostname ) {
              var segments = URI_PARSER.exec( hostname );
              if ( segments ) {
                if ( segments[ 1 ] ) {
                  uri.protocol = segments[ 1 ];
                }
                else {
                  uri.hostname = segments[ 2 ];
                  uri.port = segments[ 3 ];
                }
              }
              else {
                error( 'URI: An invalid hostname was given.' );
              }
            }
            else {
              uri.hostname = null;
              uri.port = null;
            }
          } );
        },

        setHost: function( host ) {
          return this.spawn( function( uri ) {
            uri.hostname = host;
          } );
        },

        setPort: function( port ) {
          return this.spawn( function( uri ) {
            uri.port = port;
          } );
        },

        setPath: function( path ) {
          return this.spawn( function( uri ) {
            uri.path = normalizePath( path );
          } );
        },

        joinPath: function() {
          var paths = Array.prototype.slice.call( arguments );
          return this.spawn( function( uri ) {
            var join = function( path ) {
              path = normalizePath( path );
              if ( path ) {
                uri.path = ( normalizePath( uri.path ) || '' ) + '/' + path;
              }
              else {
                throw error( 'URI: An invalid path was given.' );
              }
            };
            for ( var index = 0; index < paths.length; index++ ) {
              join( paths[ index ] );
            }
          } );
        },

        setQuery: function( query ) {
          return this.spawn( function( uri ) {
            if ( isObject( query ) ) {
              query = queryFromObject( query );
            }
            uri.query = query;
          } );
        },

        joinQuery: function( query ) {
          return this.spawn( function( uri ) {
            if ( isObject( query ) ) {
              query = queryFromObject( query );
            }
            if ( uri.query && query ) {
              uri.query = uri.query + '&' + query;
            }
            else if ( query ) {
              uri.query = query;
            }
          } );
        },

        setHash: function( hash ) {
          return this.spawn( function( uri ) {
            uri.hash = hash;
          } );
        }

      };

      var URIFactory = function( url ) {
        return new URI( url );
      };

      URIFactory.parse = function( url ) {
        var result = null;
        try {
          result = new URI( url );
        }
        catch ( exception ) {
          result = undefined;
        }
        return result;
      };

      extend(
        URIFactory,
        {
          'encodeQuery': encodeQuery,
          'encodeQuerySegment': encodeQuerySegment,
          'queryFromObject': queryFromObject,
          'objectFromQuery': objectFromQuery
        }
      );

      return URIFactory;

    }
  );
