import {
  isArray,
  isString,
  isUndefined,
  isFunction,
  isAnnotatedFunction,
  isDefined,
  isArrayLike
} from './functions/rtti';
import { map, each } from './functions/enumeration';
import { error, safely } from './functions/error';

// ============================================================================
// h1. Module

//
// example:
// unimatrix.module( 'Unimatrix' ).
//   provider( 'Http', [ 'URI', function() {
//   } )
//
// unimatrix.module( 'Boxxspring.Theme' ).
//   provider(
//     'Something',
//     [ 'Unimatrix/Uri',
//        function( uri ) {
//
//        }
//     ]
//   );
//

var modules = {};

function Module( name ) {
  this.$$name = name;
  this.$$services = {};
}

Module.prototype = {
  'constructor':  Module,

  'requires': function() {
    each( arguments, function( moduleName ) {
      if ( !isDefined( modules[ moduleName ] ) ) {
        modules[ moduleName ] = null;
      }
    } );

    return this;
  },

  'alias': function( aliasName ) {
    if ( modules[ aliasName ] && modules[ aliasName ] !== this ) {
      throw error( 'The module name', aliasName, 'is already in use.' );
    }
    modules[ aliasName ] = this;

    return this;
  },

  'service': function( serviceName, annotatedFunction ) {
    if ( isDefined( annotatedFunction ) ) {
      // note: if a service with the service name is already defined for this
      // module it will be replaced.
      this.$$services[ serviceName ] = new Service( annotatedFunction );
    }
    else {
      if ( isUndefined( this.$$services[ serviceName ] ) ) {
        throw error(
          'The module',
          this.$$name,
          'has no service named "' + serviceName + '"'
        );
      }
    }
    return this.$$services[ serviceName ];
  }
};

function module( name ) {
  var module = modules[ name ];

  if ( !isDefined( module ) ) {
    module = modules[ name ] = new Module( name );
  }
  return module;
}

// ============================================================================
// h1. Service

function Service( factoryFunction ) {
  if ( !isFunction( factoryFunction ) &&
       !isAnnotatedFunction( factoryFunction ) ) {
    throw error(
      'A service must be defined with a function or annotated function.'
    );
  }
  this.$$instance = undefined;
  this.$$factoryFunction = factoryFunction;
}

Service.prototype = {
  'constructor':  Service,

  'instance': function( localInjector ) {
    if ( isUndefined( this.$$instance ) ) {
      if ( isFunction( this.$$factoryFunction ) ) {
        this.$$instance = this.$$factoryFunction();
      }
      else {
        this.$$instance = ( localInjector || injector() ).
          invoke( this.$$factoryFunction );
      }
    }

    return this.$$instance;
  }
};

// ============================================================================
// h1. Injector

function Injector( moduleNames ) {
  this.$$modules = {};
  if ( !moduleNames ) {
    this.$$modules = modules;
  }
  else {
    this.$$modules = {};
    each(
      moduleNames,
      function( moduleName ) {
        var module = modules[ moduleName ];
        if ( !module ) {
          throw error( 'The module named', moduleName, 'is not defined.' );
        }
        this.$$modules[ moduleName ] = module;
      },
      this
    );
  }
}

Injector.prototype = {
  'get': function( moduleNameAndServiceName ) {
    var moduleName;
    var serviceName;

    if ( isString( moduleNameAndServiceName ) ) {
      moduleNameAndServiceName = moduleNameAndServiceName.split( '/' );
    }

    if ( !isArray( moduleNameAndServiceName ) ) {
      throw error(
        'The injector requires a module and service name to return a service.'
      );
    }

    if ( isDefined( moduleNameAndServiceName[ 1 ] ) ) {
      moduleName = moduleNameAndServiceName[ 0 ];
      serviceName = moduleNameAndServiceName[ 1 ];
    }
    else {
      moduleName = 'Unimatrix';
      serviceName = moduleNameAndServiceName[ 0 ];
    }

    var module = this.$$modules[ moduleName ];
    if ( !module ) {
      throw error(
        'The module named "' + moduleName +
        '" is not known to this injector.'
      );
    }

    var service = safely( function() {
      return module.service( serviceName );
    } );

    return isDefined( service ) ? service.instance( this ) : undefined;
  },

  'invoke': function( annotatedFunction ) {
    var serviceConstructor;
    var dependentServices;

    // is the annotated function an array ( annotations + function )?
    if ( isArrayLike( annotatedFunction ) ) {
      serviceConstructor = annotatedFunction[ annotatedFunction.length - 1 ];
      dependentServices = annotatedFunction.slice(
        0,
        annotatedFunction.length - 1
      );
    }

    //
    dependentServices = map(
      dependentServices,
      function( dependentServiceName ) {
        return this.get( dependentServiceName );
      },
      this
    );

    return serviceConstructor.apply( this, dependentServices );
  },

  'list': function( moduleName ) {
    var moduleObj = this.$$modules[ moduleName ];

    if ( !isDefined( moduleObj ) ) {
      throw error( 'Unknown module name' );
    }

    return Object.keys( moduleObj.$$services );
  }
};

var defaultInjector = undefined;

function injector( moduleNames ) {
  if ( moduleNames ) {
    return new Injector( moduleNames );
  }

  if ( isUndefined( defaultInjector ) ) {
    defaultInjector = new Injector();
  }

  return defaultInjector;
}

export { module, injector };
